package com.six.dochub.soapclient;

import com.six.dochub.soapclient.gen.DocumentTypeIdsType;
import com.six.dochub.soapclient.gen.GetAuthorizationTokenRequest;
import com.six.dochub.soapclient.gen.GetAuthorizationTokenResponse;
import com.six.dochub.soapclient.gen.GetDocumentRequest;
import com.six.dochub.soapclient.gen.GetDocumentResponse;
import com.six.dochub.soapclient.gen.InstrumentIdentifierType;
import com.six.dochub.soapclient.gen.IssuerIdentifierType;
import com.six.dochub.soapclient.gen.JurisdictionCodesType;
import com.six.dochub.soapclient.gen.LanguageCodesType;
import com.six.dochub.soapclient.gen.SearchDocumentRequest;
import com.six.dochub.soapclient.gen.SearchDocumentResponse;
import lombok.Getter;
import lombok.extern.slf4j.Slf4j;
import org.springframework.lang.NonNull;
import org.springframework.lang.Nullable;
import org.springframework.ws.WebServiceMessage;
import org.springframework.ws.client.core.support.WebServiceGatewaySupport;
import org.springframework.ws.soap.saaj.SaajSoapMessage;

import javax.xml.soap.MimeHeaders;
import javax.xml.soap.SOAPMessage;
import java.util.Collection;
import java.util.List;

/**
 * Provides DocHub WS Gateway support.
 *
 * @author tsc and the ARIVA.DE REG III Team
 * @since 2020-08-26.
 */
@Slf4j
@Getter
public class DocHubClient extends WebServiceGatewaySupport {
    private static final int OK_CODE = 1;

    private String authToken;

    @NonNull
    private static SearchDocumentRequest makeSearchDocumentRequest(
            final InstrumentIdentifierType instrumentIdentifier,
            final Collection<Integer> documentTypes,
            final Integer mimeType,
            final Collection<String> languages,
            final Collection<String> jurisdictions,
            final String documentSupplier,
            final IssuerIdentifierType issuerIdentifier,
            final Integer limit,
            final Integer page
    ) {
        final SearchDocumentRequest request = new SearchDocumentRequest();
        request.setInstrumentIdentifier(instrumentIdentifier);
        final DocumentTypeIdsType documentTypeIdsType = new DocumentTypeIdsType();
        documentTypeIdsType.getDocumentType().addAll(documentTypes);
        request.setDocumentTypes(documentTypeIdsType);
        request.setMimeType(mimeType);
        final LanguageCodesType languageCodesType = new LanguageCodesType();
        languageCodesType.getLanguageCode().addAll(languages);
        request.setLanguageCodes(languageCodesType);
        final JurisdictionCodesType jurisdictionCodesType = new JurisdictionCodesType();
        jurisdictionCodesType.getJurisdictionCode().addAll(jurisdictions);
        request.setJurisdictionCodes(jurisdictionCodesType);
        request.setDocumentSupplier(documentSupplier);
        request.setIssuerIdentifier(issuerIdentifier);
        request.setListLimit(limit);
        request.setListPage(page);
        return request;
    }

    /**
     * Retrieves an authorization token, using valid credentials.
     * If successful, the result is cached in {@link #authToken}.
     *
     * @param username the username
     * @param password the password
     * @return the response
     */
    public GetAuthorizationTokenResponse getAuthorizationToken(
            @NonNull final String username,
            @NonNull final String password
    ) {
        final GetAuthorizationTokenRequest request = new GetAuthorizationTokenRequest();
        request.setUsername(username);
        request.setPassword(password);

        final GetAuthorizationTokenResponse response =
                (GetAuthorizationTokenResponse) getWebServiceTemplate()
                        .marshalSendAndReceive(request);
        if ((response != null) && (response.getStatusCode() == OK_CODE)) {
            this.authToken = response.getAuthorizationToken();
        }
        return response;
    }

    /**
     * Searches for documents.
     *
     * @param instrumentIdentifier the instrument identifier
     * @param documentTypes        list of document types
     * @param mimeType             the MIME type
     * @param languages            list of languages
     * @param jurisdictions        list of jurisdictions
     * @param documentSupplier     the document supplier
     * @param issuerIdentifier     the issuer identifier
     * @param limit                the limit
     * @param page                 the result page
     * @return the response
     */
    public SearchDocumentResponse searchDocument(
            @Nullable final InstrumentIdentifierType instrumentIdentifier,
            @Nullable final List<Integer> documentTypes,
            @Nullable final Integer mimeType,
            @Nullable final List<String> languages,
            @Nullable final List<String> jurisdictions,
            @Nullable final String documentSupplier,
            @Nullable final IssuerIdentifierType issuerIdentifier,
            @Nullable final Integer limit,
            @Nullable final Integer page
    ) {
        final SearchDocumentRequest request = makeSearchDocumentRequest(
                instrumentIdentifier,
                documentTypes,
                mimeType,
                languages,
                jurisdictions,
                documentSupplier,
                issuerIdentifier,
                limit,
                page
        );

        return (SearchDocumentResponse) getWebServiceTemplate()
                .marshalSendAndReceive(request, this::addAuthToken);
    }

    /**
     * Retrieves a document.
     *
     * @param documentId  the document ID
     * @param archive     whether or not to archive
     * @param customData1 the custom data 1
     * @param customData2 the custom data 2
     * @param customData3 the custom data 3
     * @param customData4 the custom data 4
     * @param customData5 the custom data 5
     * @return the response
     */
    public GetDocumentResponse getDocument(
            @NonNull final String documentId,
            final boolean archive,
            @Nullable final String customData1,
            @Nullable final String customData2,
            @Nullable final String customData3,
            @Nullable final String customData4,
            @Nullable final String customData5
    ) {
        final GetDocumentRequest request = new GetDocumentRequest();
        request.setDocumentID(documentId);
        request.setArchive(archive);
        request.setCustomData1(customData1);
        request.setCustomData2(customData2);
        request.setCustomData3(customData3);
        request.setCustomData4(customData4);
        request.setCustomData5(customData5);

        return (GetDocumentResponse) getWebServiceTemplate()
                .marshalSendAndReceive(request, this::addAuthToken);
    }

    private void addAuthToken(final WebServiceMessage message) {
        if (message instanceof SaajSoapMessage) {
            final SaajSoapMessage soapMessage = (SaajSoapMessage) message;
            final SOAPMessage saajMessage = soapMessage.getSaajMessage();
            final MimeHeaders mimeHeaders = saajMessage.getMimeHeaders();
            mimeHeaders.addHeader("Authorization", "Bearer " + authToken);
        }
    }
}
