package com.six.dochub.soapclient;

import com.six.dochub.soapclient.gen.GetAuthorizationTokenResponse;
import com.six.dochub.soapclient.gen.GetDocumentResponse;
import com.six.dochub.soapclient.gen.InstrumentIdentifierType;
import com.six.dochub.soapclient.gen.IssuerIdentifierType;
import com.six.dochub.soapclient.gen.SearchDocumentResponse;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.ValueSource;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.test.context.ContextConfiguration;
import org.springframework.test.context.junit.jupiter.SpringExtension;
import org.springframework.test.context.support.AnnotationConfigContextLoader;
import org.springframework.ws.soap.client.SoapFaultClientException;

import java.util.List;
import java.util.Optional;

import static org.assertj.core.api.Assertions.assertThat;
import static org.assertj.core.api.Assertions.catchThrowableOfType;

@ExtendWith(SpringExtension.class)
@ContextConfiguration(classes = DocHubClientConfig.class, loader = AnnotationConfigContextLoader.class)
class DocHubClientIntegrationTest {
    private static final int OK_CODE = 1;
    private static final int INVALID_PARAMETER_CODE = 201;
    private static final int EMPTY_RESULT_CODE = 210;

    private static final int ISIN_SCHEME_TYPE = 1;
    private static final int GK_SCHEME_TYPE = 1;

    private static final String USERNAME = "my_user_account"; // TODO replace with own user account
    private static final String PASSWORD = "my_password"; // TODO replace with user password

    private static final InstrumentIdentifierType INSTRUMENT_IDENTIFIER_TYPE = new InstrumentIdentifierType();
    private static final IssuerIdentifierType ISSUER_IDENTIFIER_TYPE = new IssuerIdentifierType();

    static {
        INSTRUMENT_IDENTIFIER_TYPE.setSchemeType(ISIN_SCHEME_TYPE);
        INSTRUMENT_IDENTIFIER_TYPE.setValue("DE");

        ISSUER_IDENTIFIER_TYPE.setSchemeType(GK_SCHEME_TYPE);
        ISSUER_IDENTIFIER_TYPE.setValue("GK_DUMMY_VALUE");
    }

    @Autowired
    private DocHubClient underTest;

    private void ensureAuthorization() {
        Optional.ofNullable(underTest.getAuthToken())
                .ifPresentOrElse(t -> {  // ok, use existing token
                }, () -> underTest.getAuthorizationToken(USERNAME, PASSWORD));
    }

    @Nested
    class WhenGetAuthorizationToken {
        @Test
        void givenValidCredentials_thenGetToken() {
            final GetAuthorizationTokenResponse response = underTest.getAuthorizationToken(USERNAME, PASSWORD);
            assertThat(response.getStatusCode()).isEqualTo(OK_CODE);
            assertThat(response.getAuthorizationToken()).isNotBlank();
        }

        @Test
        void givenInvalidCredentials_thenClientError() {
            final SoapFaultClientException caught
                    = catchThrowableOfType(
                    () -> underTest.getAuthorizationToken("try", "guess"),
                    SoapFaultClientException.class
            );
            assertThat(caught).hasMessageContaining("internal error");
        }
    }

    @Nested
    class WhenSearchDocument {
        private static final int DOCUMENT_TYPE = 1;
        private static final int MIME_TYPE = 1;
        private static final String LANGUAGE = "de";
        private static final String JURISDICTION = "DE";
        private static final String DOCUMENT_SUPPLIER = "SIX pre-generating issuer";
        private static final int LIMIT = 3;
        private static final int RESULT_PAGE = 1;

        @BeforeEach
        void setUp() {
            ensureAuthorization();
        }

        @Test
        void givenValidRequest_thenGetResult() {
            final SearchDocumentResponse response = underTest.searchDocument(
                    INSTRUMENT_IDENTIFIER_TYPE,
                    List.of(DOCUMENT_TYPE),
                    MIME_TYPE,
                    List.of(LANGUAGE),
                    List.of(JURISDICTION),
                    DOCUMENT_SUPPLIER,
                    null,
                    LIMIT,
                    RESULT_PAGE
            );
            assertThat(response.getStatusCode()).isEqualTo(OK_CODE);
            assertThat(response.getTotalHits()).isPositive();
            assertThat(response.getDocuments().getDocument()).isNotEmpty();
        }

        @Test
        void givenUnknownIsin_thenGetEmptyResult() {
            final InstrumentIdentifierType instrumentIdentifier = new InstrumentIdentifierType();
            instrumentIdentifier.setSchemeType(1);
            instrumentIdentifier.setValue("DE1234XY");

            final SearchDocumentResponse response = underTest.searchDocument(
                    instrumentIdentifier,
                    List.of(DOCUMENT_TYPE),
                    MIME_TYPE,
                    List.of(LANGUAGE),
                    List.of(JURISDICTION),
                    DOCUMENT_SUPPLIER,
                    null,
                    LIMIT,
                    RESULT_PAGE
            );
            assertThat(response.getStatusCode()).isEqualTo(EMPTY_RESULT_CODE);
            assertThat(response.getTotalHits()).isZero();
            assertThat(response.getDocuments().getDocument()).isEmpty();
        }
    }

    @Nested
    class WhenGetDocument {
        private static final String
                DOCUMENT_ID =
                "acdb5cqcutb94mjszhs8h11ncer9wgc5";

        private static final String CUSTOM_DATA_1 = "Spring WS + JAXB Demo Project";
        private static final String CUSTOM_DATA_2 = "FT";
        private static final String CUSTOM_DATA_3 = "Soap V3x1";
        private static final String CUSTOM_DATA_4 = "Release 2.6.0";
        private static final String CUSTOM_DATA_5 = "Client Version 0.2.0";

        @BeforeEach
        void setUp() {
            ensureAuthorization();
        }

        @ParameterizedTest(name = "{index} shouldArchive: {0}")
        @ValueSource(strings = {"true", "false"})
        void givenValidRequest_thenRetrieveDocument(final boolean shouldArchive) {
            final GetDocumentResponse response = underTest.getDocument(
                    DOCUMENT_ID,
                    shouldArchive,
                    CUSTOM_DATA_1,
                    CUSTOM_DATA_2,
                    CUSTOM_DATA_3,
                    CUSTOM_DATA_4,
                    CUSTOM_DATA_5
            );
            assertThat(response.getStatusCode()).isEqualTo(OK_CODE);
            assertThat(response.getDocumentID()).isEqualTo(DOCUMENT_ID);
            if (shouldArchive) {
                assertThat(response.getArchiveID()).isNotBlank();
            } else {
                assertThat(response.getArchiveID()).isNullOrEmpty();
            }
            assertThat(response.getDocument()).isNotEmpty();
        }

        @ParameterizedTest(name = "{index} shouldArchive: {0}")
        @ValueSource(strings = {"true", "false"})
        void givenInvalidDocumentId(final boolean shouldArchive) {
            final String invalidDocumentId = "abcsaldjkflkasdjfsadkfj";

            final SoapFaultClientException
                    caught = catchThrowableOfType(
                    () -> underTest.getDocument(
                            invalidDocumentId,
                            shouldArchive,
                            CUSTOM_DATA_1,
                            CUSTOM_DATA_2,
                            CUSTOM_DATA_3,
                            CUSTOM_DATA_4,
                            CUSTOM_DATA_5
                    ),
                    SoapFaultClientException.class
            );
            assertThat(caught)
                    .hasMessageContaining(String.valueOf(INVALID_PARAMETER_CODE));
        }
    }
}
